from datetime import datetime
from typing import Dict, Iterator, List, Tuple

from .base import BaseAPIModel


__all__ = ['Population']


class Population(BaseAPIModel):
    """"Wrapper class for /population response

    Attributes
    ----------
    total : int
        the total population of Toontown Rewritten

    last_updated : datetime
        the time of the last population update

    population_by_district : Dict[str, int]
        a dictionary mapping district name to the district's population
    """

    __slots__ = ['total', 'last_updated', 'population_by_district']

    def __init__(self, **payload) -> None:
        self.total: int = payload.get('totalPopulation')
        self.last_updated: datetime = datetime.fromtimestamp(payload.get('lastUpdated'))
        self.population_by_district: Dict[str, int] = payload.get('populationByDistrict')

    def iteritems(self) -> Iterator[Tuple[str, int]]:
        """Return an iterator of the districts and their population"""
        yield from self.population_by_district.items()

    def keys(self) -> List[str]:
        """Returns a list of the district names"""
        return list(self.population_by_district.keys())

    def values(self) -> List[str]:
        """Returns a list of the district populations"""
        return list(self.population_by_district.values())

    def __getitem__(self, district: str) -> int:
        """Returns the population of the given district"""
        return self.population_by_district.__getitem__(district)



from typing import Dict, Iterator, List, Tuple

from .base import BaseAPIModel


__all__ = ['Doodle', 'Playground', 'District', 'Doodles']


class Doodle:
    """Wrapper class for Doodle data
    
    Attributes
    ----------
    dna : str
        the DNA that makes up the Doodle, used for rendering

    rendition : str
        the URL to a rendering of the Doodle, 256x256 .png format

    traits : List[str]
        the list of the Doodle's traits

    cost : int
        how many jellybeans the Doodle costs at the Pet Shop
    """

    __slots__ = ['dna', 'rendition', 'traits', 'cost']

    def __init__(self, *, dna, traits, cost) -> None:
        self.dna: str = dna
        self.rendition: str = f'https://rendition.toontownrewritten.com/render/{dna}/doodle/256x256.png'
        self.traits: List[str] = traits
        self.cost: int = cost


class Playground:
    """Wrapper class for Playground data
    
    Attributes
    ----------
    name : str
        the name of the playground
        
    doodles : List[Doodle]
        the list of doodles in the playground
    """

    __slots__ = ['name', 'doodles']

    def __init__(self, name, doodles) -> None:
        self.name: str = name
        self.doodles: List[Doodle] = [Doodle(**doodle) for doodle in doodles]


class District:
    """Wrapper class for District data
    
    Attributes
    ----------
    name : str
        the name of the district

    playgrounds : Dict[str, Playground]
        a dictionary mapping Playground name to Playground data
    """

    __slots__ = ['name', 'playgrounds']

    def __init__(self, name, **playgrounds) -> None:
        self.name: str = name
        self.playgrounds: Dict[str, Playground] = {pg: Playground(pg, doodles) for pg, doodles in playgrounds.items()}

    def iteritems(self) -> Iterator[Tuple[str, int]]:
        """Return an iterator of the districts and their population"""
        yield from self.playgrounds.items()

    def keys(self) -> List[str]:
        """Returns a list of the district names"""
        return list(self.playgrounds.keys())

    def values(self) -> List[str]:
        """Returns a list of the district populations"""
        return list(self.playgrounds.values())

    def __getitem__(self, district: str) -> int:
        """Returns the population of the given district"""
        return self.playgrounds.__getitem__(district)


class Doodles(BaseAPIModel):
    """"Wrapper class for /doodles response
    
    Attributes
    ----------
    districts : Dict[str, District]
        a dictionary mapping District name to District data
    """
    
    __slots__ = ['districts']

    def __init__(self, **payload) -> None:
        self.districts: Dict[str, District] = {district: District(district, **playgrounds) for district, playgrounds in payload.items()}

    def iteritems(self) -> Iterator[Tuple[str, int]]:
        """Return an iterator of the districts and their population"""
        yield from self.districts.items()

    def keys(self) -> List[str]:
        """Returns a list of the district names"""
        return list(self.districts.keys())

    def values(self) -> List[str]:
        """Returns a list of the district populations"""
        return list(self.districts.values())

    def __getitem__(self, district: str) -> int:
        """Returns the population of the given district"""
        return self.districts.__getitem__(district)


import toontown
import toontown.models


def main():
    with toontown.SyncToontownClient() as client:
        print(client.doodles())
        print(client.field_offices())
        print(client.invasions())
        # print(client.login(username, password))
        print(client.population())


main()



import functools
from typing import Any, Dict

import yaml


__all__ = ['get']


with open('config.yaml', 'r') as file:
    _config: Dict[str, Any] = yaml.load(file, Loader=yaml.FullLoader)


def get(*keys: str) -> Any:
    """Recursively get from config
    
    EX:
    >>> # config.yaml
    >>> http:
    >>>  base: https://base.com
    >>> ...
    >>> # python3
    >>> config.get('http', 'base')
    https://base.com
    """
    return functools.reduce(lambda c, k: c.get(k, {}), keys, _config)



from ..exceptions import FailedResponse


class BaseAPIModel:
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)

        if kwargs.get('error', None):
            """Sometimes the server will send an error field when there is no cached response"""
            raise FailedResponse(kwargs['error'])
            
        return instance


import asyncio
import logging
import time
from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Union
)

import aiohttp
import requests
import requests.exceptions

from . import config
from .exceptions import *


logger = logging.getLogger(__name__)


Session = Union[aiohttp.ClientSession, requests.Session]

BASE: str = config.get('http', 'base')
BASE_HEADERS: Dict[str, Any] = config.get('http', 'base_headers')
LOGIN_HEADERS: Dict[str, Any] = config.get('http', 'login_headers')


class Route:
    def __init__(self, method: str, path: str, **parameters: Any) -> None:
        self.method = method
        self.path = path
        self.parameters = parameters
        self.url = BASE + path

    @property
    def headers(self) -> Dict[str, Any]:
        return LOGIN_HEADERS if self.path == '/login' else BASE_HEADERS


class BaseHTTPClient(ABC):
    @abstractmethod
    def __init__(self) -> None:
        self._session: Session = None

    @abstractmethod
    def connect(self) -> None: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def request(self, route: Route) -> Any: ...


class SyncHTTPClient(BaseHTTPClient):
    def __init__(self) -> None:
        self._session: requests.Session = None
        self._is_closed = True

    @property
    def is_closed(self) -> bool:
        return self._is_closed

    @is_closed.setter
    def is_closed(self, value: bool) -> None:
        self._is_closed = bool(value)

    def connect(self) -> None:
        self._session = requests.Session()
        self._is_closed = False

    def close(self) -> None:
        self._session.close()
        self._is_closed = True

    def request(self, route: Route) -> Any:
        if self.is_closed:
            raise SessionNotConnected

        method = route.method
        url = route.url
        headers = route.headers
        params = route.parameters

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                with self._session.request(method, url, params=params, headers=headers) as response:
                    response.raise_for_status()

                    data = response.json()
                    status = response.status_code

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        time.sleep(1 + tries * 2)
                        continue

            except requests.exceptions.HTTPError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    time.sleep(1 + tries * 2)
                    continue

                logger.info('Exhausted attempts for {0} request: {1}'.format(method, url))
                raise


class AsyncHTTPClient(BaseHTTPClient):
    def __init__(self) -> None:
        self._session: aiohttp.ClientSession = None

    async def connect(self) -> None:
        self._session = aiohttp.ClientSession(raise_for_status=True)

    async def close(self) -> None:
        self._session.close()

    async def request(self, route: Route) -> Any:
        if self._session.closed:
            raise SessionNotConnected

        method = route.method
        url = route.url
        headers = route.headers
        params = route.parameters

        for tries in range(5):
            try:
                logger.info('Attempting {0} request #{1}: {2}'.format(method, tries+1, url))

                async with self._session.request(method, url, params=params, headers=headers) as response:
                    data = await response.json()
                    status = response.status

                    if 300 > status >= 200:
                        return data

                    if status in {500, 502, 504}:
                        await asyncio.sleep(1 + tries * 2)
                        continue

            except aiohttp.ClientResponseError as e:
                message = str(e)
                exc_info = type(e), e, e.__traceback__
                logger.error(message, exc_info=exc_info)
                raise

            except OSError as e:
                if tries < 4 and e.errno in {54, 10054}:
                    await asyncio.sleep(1 + tries * 2)
                    continue
                raise



from dataclasses import dataclass

from .base import BaseAPIModel


__all__ = ['Login', 'FailedLogin', 'PartialLogin', 'DelayedLogin', 'SuccessfulLogin']


class Login(BaseAPIModel):
    """"Wrapper class for /login response

    Returns
    -------
    SuccessfulLogin
        you can now log into Toontown Rewritten with the provided gameserver and cookie

    DelayedLogin
        you are in the login queue and must poll again with the given queueToken

    FailedLogin
        your account is either not verified or is disabled

    PartialLogin
        your account must be authenticated with ToonGuard
    """
    
    def __new__(cls, *args, **kwargs):
        super().__new__(cls, *args, **kwargs)
        
        success = kwargs.get('success')
        if success == 'true':
            return SuccessfulLogin(kwargs.get('gameserver'), kwargs.get('cookie'))
        elif success == 'delayed':
            return DelayedLogin(kwargs.get('eta'), kwargs.get('position'), kwargs.get('queueToken'))

        banner = kwargs.get('banner')

        if success == 'false':
            return FailedLogin(banner)
        elif success == 'partial':
            return PartialLogin(banner, kwargs.get('responseToken'))


@dataclass
class FailedLogin:
    """Login response if your account was not verified or was disabled
    
    Attributes
    ----------
    banner : str
        the server's response to why the login was failed
    """
    banner: str


@dataclass
class PartialLogin:
    """Login response if your account needs to be authenticated with ToonGuard
    
    Attributes
    ----------
    banner : str
        the server's response to why the login is partial

    response_token : str
        the token to verify that the ToonGuard code was sent by you    
    """
    banner: str
    response_token: str


@dataclass
class DelayedLogin:
    """Login response if you are in a queue to log in
    
    Attributes
    ----------
    eta : int
        the estimated time in seconds of when you can expect to log in
        
    position : int
        your position in the queue
        
    queue_token : str
        the token given in the initial delayed response to update your position in the queue
    """
    eta: int
    position: int
    queue_token: str


@dataclass
class SuccessfulLogin:
    """Login response if you are successfully logged into Toontown Rewritten
    
    Attributes
    ----------
    gameserver : str
        Toontown Rewritten's server host used to connect to the game server
        
    cookie : str
        your unique identifier for your current session, DO NOT SHARE
    """
    gameserver: str
    cookie: str



from datetime import datetime
from typing import List

from .base import BaseAPIModel


__all__ = ['Invasion', 'Invasions']


class Invasion:
    """Wrapper class for invasion data
    
    Attributes
    ----------
    district : str
        the district the invasion is in

    as_of : datetime
        when the invasion started

    type : str
        the Cog type of the invasion

    progress : int
        how many cogs were defeated in this invasion

    total : int
        how many cogs that are invading

    is_mega_invasion : bool
        whether or not this is a mega invasion
    """
    def __init__(self, district, **payload) -> None:
        self.district: str = district
        self.as_of = datetime.fromtimestamp(payload.pop('asOf'))
        self.type: str = payload.pop('type')

        progress, total = payload.pop('progress').split('/')
        self.progress = int(progress)
        self.total = int(total)

    @property
    def is_mega_invasion(self) -> bool:
        return self.total == 1000000


class Invasions(BaseAPIModel):
    """"Wrapper class for /invasions response

    Attributes
    ----------
    last_updated : datetime
        the time when the invasions were last updated

    invasions : List[Invasion]
        the list of invasion data for each district
    """

    __slots__ = ['last_updated', 'invasions']
    
    def __init__(self, **payload) -> None:
        self.last_updated = datetime.fromtimestamp(payload.pop('lastUpdated'))
        
        invasions = payload.pop('invasions')
        self.invasions: List[Invasion] = [Invasion(district, **props) for district, props in invasions.items()]